/*
 *  drivers/mtd/mx2_nand.c
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   MX21 board, It supports 8MiB cards.
 *
 * Author: MontaVista Software, Inc.  source@mvista.com
 *
 * Copyright (C) 2000 Motorola, Inc.
 *
 * 2004 (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 */

#include <linux/slab.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/arch/gpio.h>

#ifdef CONFIG_MTD_NAND_ECC
#include <linux/mtd/nand_ecc.h>
#endif

#include "mx2_nand.h"

#ifdef CONFIG_MTD_CMDLINE_PARTS
const char *mx2_probes[] = { "cmdlinepart", NULL };
#endif

extern int nand_correct_data(struct mtd_info *mtd, u_char * dat,
			     u_char * read_ecc, u_char * calc_ecc);

static u32 *mab;
static u32 mab_counter;
static u32 mx2_cmd;
static u32 ic = 0;
static u32 mx2_page;
static u32 mx2_nand_status = 0;
static struct mtd_info *mx2_mtd = NULL;

#define NAND_TIMEOUT HZ*5
#ifdef  CONFIG_MX2_16_BIT_NAND
#define ADDRCYCLE  3
#else
#define ADDRCYCLE  4
#endif

void
WAIT(void)
{
	unsigned long timeout;
	timeout = jiffies + NAND_TIMEOUT;
	while (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT)) {
		if (time_after(jiffies, timeout)) {
			printk("timeout waiting for Nand command complete");
			return;
		}
		if (in_interrupt()) {
			mdelay(10);
		} else {
			schedule_timeout(1);
		}
	}

}

void
Preset(void)
{

	NFC_CONFIGURATION = 0x0002;
	NFC_ULOCK_START_BLK = 0;
	NFC_ULOCK_END_BLK = NAND_FLASH_BLOCK_NUMBER-1;
	NFC_NF_WR_PROT = 0x0004;
	NFC_NF_CONFIG1 = 0x000A;
	NFC_RAM_BUF_ADDR = 0x3;
}

u32
NAND_Command(u16 command)
{
	NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_INT;

	NFC_NAND_FLASH_CMD = command;
	NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_FCMD;
	NFC_NF_CONFIG2 |= NAND_FLASH_CONFIG2_FCMD;

	WAIT();
	if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT))
		return ERR_TIME_OUT;

	return 0;
}

u32
NAND_SendAddress(u32 addr, int cycle)
{
	int i = 0;
	for (i = 0; i < cycle; i++) {
		NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_INT;
		if (cycle == ADDRCYCLE && i > 0) {
			NFC_NAND_FLASH_ADDR = (addr >> (i * 8 + 1)) & 0xFF;
		} else {
			NFC_NAND_FLASH_ADDR = (addr >> i * 8) & 0xFF;
		}
		NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_FADD;
		NFC_NF_CONFIG2 |= NAND_FLASH_CONFIG2_FADD;

		WAIT();
		if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT))
			return ERR_TIME_OUT;
	}
	return 0;
}

static u_char
mx2_nand_read_byte(struct mtd_info *mtd)
{
	char b;

	b = (*mab) >> ic * 8;
	ic++;
#ifdef  CONFIG_MX2_16_BIT_NAND
	if (mx2_cmd == NAND_CMD_READID || mx2_cmd == NAND_CMD_STATUS)
		ic++;
#endif

	if (ic == 4) {
		ic = 0;
		mab++;
	}
	return b;
}

static void
mx2_nand_command(struct mtd_info *mtd, unsigned command, int column,
		 int page_addr)
{
	int num = 0;
	int i;
	char sc;
	int address;

	mx2_cmd = command;
	mx2_page = page_addr;
	Preset();
	if (command == NAND_CMD_PAGE_PROG_CONFIRM_TRUE) {
		NFC_RAM_BUF_ADDR = 0x3;
		NFC_NF_CONFIG2 = NAND_FLASH_CONFIG2_FDI;
		WAIT();
		if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT)) {
			return;
		}
	}

	if (command == NAND_CMD_ERASE1 || command == NAND_CMD_PAGE_PROG) {
		NFC_BLK_ADD_LOCK = page_addr;
	}
	if (command == NAND_CMD_PAGE_PROG) {
		NFC_BLK_ADD_LOCK =
		    page_addr * NAND_FLASH_PAGE_MAIN_SIZE /
		    NAND_FLASH_PAGE_PER_BLOCK;
	}

	if (command == NAND_CMD_READOOB) {
		NAND_Command(NAND_CMD_READ1);
	} else {
		NAND_Command(command);
	}

	if (command == NAND_CMD_ERASE1) {
		address = page_addr;
	} else {
		address = page_addr * NAND_FLASH_PAGE_MAIN_SIZE;
	}

	switch (command) {
	case NAND_CMD_READ1:
	case NAND_CMD_READ0:
	case NAND_CMD_PAGE_PROG:
	case NAND_CMD_READOOB:
#ifdef CONFIG_MX2_16_BIT_NAND
		num = 3;
#else
		num = 4;
#endif
		break;
	case NAND_CMD_READ_ID:
		num = 1;
		break;
	case NAND_CMD_ERASE1:
#ifdef CONFIG_MX2_16_BIT_NAND
		num = 2;
#else
		num = 3;
#endif
		break;
	case NAND_CMD_STATUS:
	case NAND_CMD_PAGE_PROG_CONFIRM_TRUE:
	case NAND_CMD_BLOCK_ERASE_CONFIRM:
		num = 0;
		break;
	}
	if (num > 0) {
		NAND_SendAddress(address, num);

	}

	NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_INT;

	switch (command) {
	case NAND_CMD_READOOB:
		CONFIG2_SET_FDO_PAGE_OUT();
		WAIT();
		if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT))
			return;

		break;
	case NAND_CMD_READ_ID:
		CONFIG2_SET_FDO_ID();
		WAIT();
		if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT))
			return;

		break;

	case NAND_CMD_STATUS:
	case 0x7a:
		CONFIG2_SET_FDO_ID();
		WAIT();
		if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT))
			return;

		break;
	case NAND_CMD_ERASE1:
		break;
	case NAND_CMD_PAGE_PROG_CONFIRM_TRUE:
	case NAND_CMD_BLOCK_ERASE_CONFIRM:
		break;
	}

	mab = (u32 *) (NFC_MAB_BASE(3));
	mab_counter = 0;
	ic = 0;

	if (command == NAND_CMD_READOOB) {
		mab = (u32 *) (NFC_SAB_BASE(3));
		for (i = 1; i < column; i++) {
			sc = mx2_nand_read_byte(mtd);
		}

	}

}

static void
mx2_nand_read_buf(struct mtd_info *mtd, u_char * buf, int len)
{
	int i;
	u32 *cp = (u32 *) buf;

	if (mx2_cmd == NAND_CMD_READ0) {
		switch (mx2_nand_status) {
		case READBUF:
			mx2_nand_command(mtd, mx2_cmd, 0, ++mx2_page);
			break;
		case READCMD:
			mx2_nand_status = READBUF;
			break;
		}
	}
	mab = (u32 *) (NFC_MAB_BASE(3));
	mab_counter = 0;
	NFC_NF_CONFIG2 &= ~NAND_FLASH_CONFIG2_INT;
	CONFIG2_SET_FDO_PAGE_OUT();
	WAIT();
	if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT)) {
		return;
	}

	for (i = 0; i < len / sizeof (u32); i++) {
		*cp++ = *mab++;
	}

}

static void
mx2_nand_write_buf(struct mtd_info *mtd, const u_char * buf, int len)
{
	int i;
	u32 *cp = (u32 *) buf;
	u32 *bp = (u32 *) (NFC_MAB_BASE(3));

	if (len == 16) {
		bp = (u32 *) (NFC_SAB_BASE(3));
	}

	for (i = 0; i < len / sizeof (u32); i++) {
		*bp++ = *cp++;
	}

}

static int
mx2_nand_verify_buf(struct mtd_info *mtd, const u_char * buf, int len)
{
	int i;
	u32 *cp = (u32 *) buf;
	u32 *bp = (u32 *) (NFC_MAB_BASE(3));

	CONFIG2_SET_FDO_PAGE_OUT();
	WAIT();
	if (!(NFC_NF_CONFIG2 & NAND_FLASH_CONFIG2_INT)) {
		return ERR_TIME_OUT;
	}

	for (i = 0; i < len / sizeof (u32); i++) {
		if (*cp++ != *bp++) {
			return i * 4;
		}
	}
	return 0;
}

void
mx2_enable_hwecc(struct mtd_info *mtd, int mode)
{
}

void
mx2_calculate_ecc(struct mtd_info *mtd, const u_char * dat, u_char * ecc_code)
{
	static u32 *sab, *ecc;

	sab = (u32 *) (NFC_SAB_BASE(3));
	ecc = (u32 *) (ecc_code);
	*ecc = *sab;
}

static void
mx2_hwcontrol(struct mtd_info *mtd, int cmd)
{

	mx2_nand_status = cmd;

	if (cmd == NAND_CTL_SETNCE)
		mx2_nand_status = READCMD;
	else if (cmd == NAND_CTL_CLRNCE)
		mx2_nand_status = 0;
}

const static struct mtd_partition partition_info32m[] = {

      {name:"ipl",
	      offset:0x02000000,
      size:0x04000},

      {name:"spl",
	      offset:MTDPART_OFS_APPEND,
      size:0x100000},

      {name:"kernel",
	      offset:MTDPART_OFS_APPEND,
      size:0x200000},

      {name:"RootDisk",
	      offset:MTDPART_OFS_APPEND,
      size:MTDPART_SIZ_FULL}

};

const static struct mtd_partition partition_info64m[] = {

      {name:"ipl",
	      offset:0x02000000,
      size:0x04000},

      {name:"spl",
	      offset:MTDPART_OFS_APPEND,
      size:0x100000},

      {name:"kernel",
	      offset:MTDPART_OFS_APPEND,
      size:0x200000},

      {name:"RootDisk",
	      offset:MTDPART_OFS_APPEND,
      size:MTDPART_SIZ_FULL}

};

#define NUM_PARTITIONS32M 4
#define NUM_PARTITIONS64M 4

/**
 * Main initialization routine
 *
 * This is first function called by linux system. It will init data and call nand_scan to
 * search system nand flash
 *
 */
int __init
mx2_init(void)
{
	struct nand_chip *this;
	int retval = 0;
#ifdef  CONFIG_MX2_16_BIT_NAND
	SYS_FMCR |= 1 << 4;
#endif

#ifdef CONFIG_MX2_16_BIT_NAND
	retval =
	    mx2_register_gpios(PORT_E,
			       (1 << 23) | (1 << 24) | (1 << 25) | (1 << 27) |
			       (1 << 28) | (1 << 29) | (1 << 30) | (1 << 31),
			       SECONDARY);
#endif

	/* Allocate memory for MTD device structure and private data */
	mx2_mtd = kmalloc(sizeof (struct mtd_info) + sizeof (struct nand_chip),
			  GFP_KERNEL);
	if (!mx2_mtd) {
		printk("Unable to allocate MX2 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	mx2_mtd->oobinfo.useecc = 1;
	/* Get pointer to private data */
	this = (struct nand_chip *) (&mx2_mtd[1]);

	/* Initialize structures */
	memset((char *) mx2_mtd, 0, sizeof (struct mtd_info));
	memset((char *) this, 0, sizeof (struct nand_chip));

	/* Link the private data with the MTD structure */
	mx2_mtd->priv = this;

	this->cmdfunc = mx2_nand_command;
	this->hwcontrol = mx2_hwcontrol;
	this->read_byte = mx2_nand_read_byte;
	this->read_buf = mx2_nand_read_buf;
	this->write_buf = mx2_nand_write_buf;
	this->verify_buf = mx2_nand_verify_buf;
	this->select_chip = NULL;
	this->enable_hwecc = mx2_enable_hwecc;
	this->calculate_ecc = mx2_calculate_ecc;
	this->eccmode = NAND_ECC_HW3_512;
	this->correct_data = nand_correct_data;
	/*
	 * Set GPIO Port E control register so that the pins are configured
	 * to be outputs for controlling the NAND flash.
	 */
	/* Set address of NAND IO lines */

	/* Set address of hardware control function */

	CRM_PCDR = (CRM_PCDR & 0xFFFF0FFF) | 0x7000;
	/* Scan to find existence of the device */
	if (nand_scan(mx2_mtd, 1)) {
		kfree(mx2_mtd);
		return -ENXIO;
	}

	/* Allocate memory for internal data buffer */
	this->data_buf =
	    kmalloc(sizeof (u_char) * (mx2_mtd->oobblock + mx2_mtd->oobsize),
		    GFP_KERNEL);
	if (!this->data_buf) {
		printk("Unable to allocate NAND data buffer for Mx2.\n");
		kfree(mx2_mtd);
		return -ENOMEM;
	}

	/* Register the partitions */

#ifdef CONFIG_MTD_CMDLINE_PARTS
	{
		int mtd_parts_nb = 0;
		struct mtd_partition *mtd_parts = 0;
		mx2_mtd->name = "mx2-nand";
		mtd_parts_nb =
		    parse_mtd_partitions(mx2_mtd, mx2_probes, &mtd_parts, 0);
		if (mtd_parts_nb > 0)
			add_mtd_partitions(mx2_mtd, mtd_parts, mtd_parts_nb);
		else
			add_mtd_device(mx2_mtd);
	}
#else				/* ifdef CONFIG_MTD_CMDLINE_PARTS */
	switch (mx2_mtd->size) {
	case 0x02000000:
		add_mtd_partitions(mx2_mtd, partition_info32m,
				   NUM_PARTITIONS32M);
		break;
	case 0x04000000:
		add_mtd_partitions(mx2_mtd, partition_info64m,
				   NUM_PARTITIONS64M);
		break;
	default:{
			printk("Unsupported SmartMedia device\n");
			kfree(this->data_buf);
			return -ENXIO;
		}
	}
#endif				/* ifdef CONFIG_MTD_CMDLINE_PARTS */

	/* Return happy */
	return 0;
}

module_init(mx2_init);

/**
 * Clean up routine
 */
#ifdef MODULE
static void __exit
mx2_cleanup(void)
{
	struct nand_chip *this = (struct nand_chip *) &mx2_mtd[1];

	/* Unregister the device */
	del_mtd_device(mx2_mtd);

	/* Free internal data buffer */
	kfree(this->data_buf);

	/* Free the MTD device structure */
	kfree(mx2_mtd);

}

module_exit(mx2_cleanup);
#endif

MODULE_LICENSE("GPL");
