/*
 *
 *  85xx_list.c, version 1.0
 *  Handles list management for extended chaining mode (striding)
 *
 *  Based on hhnet_dma from MontaVista
 *
 *  This software may be used and distributed according to the terms of
 *  the GNU Public License, Version 2, incorporated herein by reference.
 *
 *  Contact:  <kumar.gala@motorola.com>
 *
 */

#include "85xx_dma_priv.h"
#include "85xx_dma.h"

int
mpc85xx_dma_alloc_list_desc(struct dma_list_desc **list_desc, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	unsigned long flags;

	Debug((PRINT_PREFIX "mpc85xx_dma_alloc_list_desc\n"));

	if ((list_desc == NULL) || (ch == NULL)) {
		Debug((PRINT_PREFIX "mpc85xx_dma_alloc_list_desc - INVAL\n"));
		return -EINVAL;
	}

	spin_lock_irqsave(&ch->lock, flags);
	if (list_empty(&ch->free_list)) {
		spin_unlock_irqrestore(&ch->lock, flags);
		Debug((PRINT_PREFIX "mpc85xx_dma_alloc_list_desc - NOMEM\n"));
		return -ENOMEM;
	} else {
		*list_desc =
		    list_entry(ch->free_list.next, struct dma_list_desc, list);
		list_del(&(*list_desc)->list);
		INIT_LIST_HEAD(&(*list_desc)->link);
		(*list_desc)->next_list = 0;
		(*list_desc)->first_link = 0;
		(*list_desc)->src_stride = 0;
		(*list_desc)->dst_stride = 0;
	}
	spin_unlock_irqrestore(&ch->lock, flags);

	return 0;
}

void
mpc85xx_dma_append_list_desc(struct list_head *list, struct list_head *head)
{
	Debug((PRINT_PREFIX "mpc85xx_dma_append_list_desc\n"));

	if (list_empty(list))
		return;

	if (!list_empty(head)) {
		struct dma_list_desc *end;
		struct dma_list_desc *qlast;
		u32 temp;

		end = list_entry(list->prev, struct dma_list_desc, list);
		qlast = list_entry(head->prev, struct dma_list_desc, list);

		temp = end->next_list;
		end->next_list = qlast->next_list;
		qlast->next_list = temp;
	}

	list_splice(list, head->prev);
}

int
mpc85xx_dma_set_striding_info(void *dma_ch, u16 src_size, u16 src_dist,
			      u16 dst_size, u16 dst_dist)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	Debug((PRINT_PREFIX "mpc85xx_dma_set_striding_info\n"));

	if (ch == NULL) {
		Debug((PRINT_PREFIX
		       "mpc85xx_dma_set_striding_info (ch = NULL) - INVAL\n"));
		return -EINVAL;
	}

	if ((ch->mode & DMA_MODE_STRIDING) != DMA_MODE_STRIDING) {
		Debug((PRINT_PREFIX
		       "mpc85xx_dma_set_striding_info (mode != striding) - INVAL\n"));
		return -EINVAL;
	}

	if (ch->curr_list != NULL) {
		if (!list_empty(&ch->curr_list->link)) {
			struct dma_link_desc *prev =
			    list_entry(ch->curr_list->link.prev,
				       struct dma_link_desc, pub.list);
			LIST_HEAD(head);

			prev->next_link = DMA_LINK_DESC_EOLND;

			list_add(&ch->curr_list->list, &head);
			mpc85xx_dma_append_list_desc(&head, &ch->pending);
			mpc85xx_dma_alloc_list_desc(&ch->curr_list, ch);
		}
	} else {
		mpc85xx_dma_alloc_list_desc(&ch->curr_list, ch);
	}

	ch->curr_list->src_stride =
	    ((src_size & 0xfff) << 12) | (src_dist & 0xfff);
	ch->curr_list->dst_stride =
	    ((dst_size & 0xfff) << 12) | (dst_dist & 0xfff);
	ch->curr_list->next_list = ch->curr_list->paddr | DMA_CLNDAR_GLEN;

	return 0;
}

static inline void
__mpc85xx_dma_free_list_desc(struct dma_list_desc *list_desc, void *dma_ch)
{
	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	unsigned int flags;

	Debug((PRINT_PREFIX "__mpc85xx_dma_free_list_desc\n"));

	spin_lock_irqsave(&ch->lock, flags);
	list_add(&list_desc->list, &ch->free_list);
	spin_unlock_irqrestore(&ch->lock, flags);
}

static inline void
__mpc85xx_dma_free_list(struct list_head *head, void *dma_ch)
{

	struct dma_channel *ch = (struct dma_channel *) dma_ch;
	unsigned long flags;

	Debug((PRINT_PREFIX "__mpc85xx_dma_free_list\n"));

	spin_lock_irqsave(&ch->lock, flags);
	list_splice(head, &ch->free_list);
	spin_unlock_irqrestore(&ch->lock, flags);
}

void
mpc85xx_dma_striding_bh(void *data)
{
	struct dma_channel *ch = (struct dma_channel *) data;
	struct list_head head;
	struct list_head *cur;
	struct list_head *cur_link;
	struct dma_link_desc *p_linkdesc;
	struct dma_list_desc *p_listdesc;
	unsigned int flags;
	int stat = 0;

	Debug((PRINT_PREFIX "mpc85xx_dma_striding_bh\n"));

	spin_lock_irqsave(&ch->lock, flags);

	if (list_empty(&ch->complete)) {
		spin_unlock_irqrestore(&ch->lock, flags);
		Debug((PRINT_PREFIX "Complete list empty\n"));
		return;
	}

	list_add(&head, &ch->complete);
	list_del_init(&ch->complete);

	spin_unlock_irqrestore(&ch->lock, flags);

	list_for_each(cur, &head) {
		p_listdesc = list_entry(cur, struct dma_list_desc, list);
		Debug((PRINT_PREFIX "processing 0x%08x\n",
		       (unsigned int) p_listdesc));

		if (p_listdesc->next_list & DMA_LIST_DESC_EOLSD) {
			Debug((PRINT_PREFIX "EOLSD Set\n"));
			if (p_listdesc->status & DMA_ERROR)
				stat = -EIO;
			else
				stat = 0;
		}

		list_for_each(cur_link, &p_listdesc->link) {
			p_linkdesc =
			    list_entry(cur_link, struct dma_link_desc,
				       pub.list);
			Debug((PRINT_PREFIX "processing 0x%08x\n",
			       (unsigned int) p_linkdesc));

			p_linkdesc->pub.status = stat;

			if (p_linkdesc->pub.callback)
				p_linkdesc->pub.callback((void *) &p_linkdesc->
							 pub);
		}
	}
	__mpc85xx_dma_free_list(&head, (void *) ch);
}

EXPORT_SYMBOL(mpc85xx_dma_alloc_list_desc);
EXPORT_SYMBOL(mpc85xx_dma_append_list_desc);
EXPORT_SYMBOL(mpc85xx_dma_set_striding_info);
EXPORT_SYMBOL(mpc85xx_dma_striding_bh);
