/*
 * Handle mapping of the flash on the ATS1460
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>


/* Main Flash - Micron 28F128J3, 16MB (64 uniform 128K sectors)
 * This device will ALWAYS exist
 */
static struct mtd_info *main_mtd;

#define MAIN_FLASH_ADDR 0xFF000000
#define MAIN_FLASH_SIZE 0x01000000

static struct map_info ats1460_main_map = {
	.name = "ATS1460_MAIN_FLASH",
	.size = MAIN_FLASH_SIZE,
	.buswidth = 2,
	.phys = MAIN_FLASH_ADDR,
};

/* FF000000 - FF01FFFF is boot strap vector - Don't mess with this!!!! */

#define APP_PARTITION_OFFSET        (0x00020000) /* FF020000 - FFBFFFFF */
#define APP_PARTITION_SIZE          (0x00BE0000) /* 12MB - 128K */

#define ROOTFS_PARTITION_OFFSET     (0x00C00000) /* FFC00000 - FFEFFFFF */
#define ROOTFS_PARTITION_SIZE       (0x00300000) /* 3MB */

#define UBOOT_PARTITION_OFFSET      (0x00F00000) /* FFF00000 - FFF3FFFF */
#define UBOOT_PARTITION_SIZE        (0x00040000) /* 256K */

/* FFF40000 - FFF60000 is UBoot Config Data */
#define UBOOT_CFG_PARTITION_OFFSET  (0x00F40000) /* FFF40000 - FFFFFFFF */
#define UBOOT_CFG_PARTITION_SIZE    (0x00020000) /* 128K */ 

#define KERN_PARTITION_OFFSET       (0x00F60000) /* FFF60000 - FFFFFFFF */
#define KERN_PARTITION_SIZE         (0x000A0000) /* 640K */ 

/*
 * MTD 'PARTITIONING' STUFF 
 */
static struct mtd_partition main_partitions[] = {
    { .name =   "ATS1460 Bootloader", 
      .offset = UBOOT_PARTITION_OFFSET, 
      .size =   UBOOT_PARTITION_SIZE            },
    { .name =   "ATS1460 Kernel", 
      .offset = KERN_PARTITION_OFFSET, 
      .size =   KERN_PARTITION_SIZE             },
    { .name =   "ATS1460 Root File System", 
      .offset = ROOTFS_PARTITION_OFFSET, 
      .size =   ROOTFS_PARTITION_SIZE           },
    { .name =   "ATS1460 Applications", 
      .offset = APP_PARTITION_OFFSET, 
      .size =   APP_PARTITION_SIZE              },
    { .name =   "ATS1460 VPD", 
      .offset = UBOOT_CFG_PARTITION_OFFSET, 
      .size =   UBOOT_CFG_PARTITION_SIZE        },
};

#define NUM_MAIN_PARTITIONS         (sizeof(main_partitions)/sizeof(struct mtd_partition))

/* Aux Flash - Micron 28F128J3, 16MB (64 uniform 128K sectors)
 * This device may not exist
 */
static struct mtd_info *aux_mtd;

#define AUX_FLASH_ADDR 0xFE000000
#define AUX_FLASH_SIZE 0x01000000

static struct map_info ats1460_aux_map = {
	.name = "ATS1460_AUX_FLASH",
	.size = AUX_FLASH_SIZE,
	.buswidth = 2,
	.phys = AUX_FLASH_ADDR,
};

#define AUX_PARTITION_OFFSET        (0x00000000) /* FE000000 - FEFFFFFF */
#define AUX_PARTITION_SIZE          (0x01000000) /* 16 MB */

/*
 * MTD 'PARTITIONING' STUFF 
 */
static struct mtd_partition aux_partitions[] =  {
    { .name =   "ATS1460 Auxiliary", 
      .offset = AUX_PARTITION_OFFSET, 
      .size =   AUX_PARTITION_SIZE              },
};

#define NUM_AUX_PARTITIONS         (sizeof(aux_partitions)/sizeof(struct mtd_partition))

int __init init_ats1460(void)
{
	int retval = 0;

	printk(KERN_NOTICE "ATS1460 main flash device: %x at %x\n", MAIN_FLASH_SIZE, MAIN_FLASH_ADDR);

	ats1460_main_map.virt = (unsigned long)ioremap(MAIN_FLASH_ADDR, MAIN_FLASH_SIZE);

	if (!ats1460_main_map.virt) {
		printk("Failed to ioremap\n");
		return -EIO;
	}

	simple_map_init(&ats1460_main_map);

	main_mtd = do_map_probe("cfi_probe", &ats1460_main_map);

	if (main_mtd) {

		main_mtd->owner = THIS_MODULE;

		retval = add_mtd_partitions(main_mtd, main_partitions, NUM_MAIN_PARTITIONS );
        
        if (!retval) {
	        printk(KERN_NOTICE "ATS1460 aux flash device: %x at %x\n", AUX_FLASH_SIZE, AUX_FLASH_ADDR);

	        ats1460_aux_map.virt = (unsigned long)ioremap(AUX_FLASH_ADDR, AUX_FLASH_SIZE);

        	if (!ats1460_aux_map.virt) {
		        printk("Failed to ioremap\n");
		        return -EIO;
	        }

        	simple_map_init(&ats1460_aux_map);

        	aux_mtd = do_map_probe("cfi_probe", &ats1460_aux_map);

        	if (aux_mtd) {
		        aux_mtd->owner = THIS_MODULE;
        		add_mtd_partitions(aux_mtd, aux_partitions, NUM_AUX_PARTITIONS );
	        }
        }
        return retval;
	}

	return -ENXIO;
}

static void __exit cleanup_ats1460(void)
{
	if (main_mtd) {
		del_mtd_device(main_mtd);
		map_destroy(main_mtd);
	}
	if (ats1460_main_map.virt) {
		iounmap((void *)ats1460_main_map.virt);
		ats1460_main_map.virt = 0;
	}
	if (aux_mtd) {
		del_mtd_device(aux_mtd);
		map_destroy(aux_mtd);
	}
	if (ats1460_aux_map.virt) {
		iounmap((void *)ats1460_aux_map.virt);
		ats1460_aux_map.virt = 0;
	}
}

module_init(init_ats1460);
module_exit(cleanup_ats1460);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("LVL7 Systems.");
MODULE_DESCRIPTION("MTD map driver for ATS1460 boards");
