/*
 * Handle mapping of the flash on the CXE 2130
 */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/io.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/map.h>
#include <linux/mtd/partitions.h>
#include <asm/8xx_immap.h>


#define WINDOW_ADDR 0xf8000000
#define WINDOW_SIZE 0x01000000

#define CFG_IMMR    0xFF000000

static struct mtd_info *mymtd;
static void cxe_2130_set_vpp(struct map_info *not_used, int on);

static struct map_info cxe_2130_map = {
	.name = "CXE-2130",
	.size = WINDOW_SIZE,
	.buswidth = 4,
	.set_vpp = cxe_2130_set_vpp,
	.phys = WINDOW_ADDR,
};


/* 
 * MAP DRIVER STUFF
 */
// - Flash / Partition sizing (sizes must be multiplse of 256K)

#define MAX_SIZE_KiB             16384
// #define BOOT_PARTITION_SIZE_KiB    256	
/* On RD1100 & RD2130, contains miniboot, u-boot, u-boot environment - 3 sectors. Miniboot is replaced by just U-boot, to be consistent with other platforms. */
#define KERN_PARTITION_SIZE_KiB   3584
#define BOOT_PARTITION_SIZE_KiB    512	
// #define APP_PARTITION_SIZE_KiB   11264
#define APP_PARTITION_SIZE_KiB   11008
#define DATA_PARTITION_SIZE_KiB   1024
#define EEPROM_PARTITION_SIZE_KiB  256

#define NUM_PARTITIONS 5



/*
 * MTD 'PARTITIONING' STUFF 
 */
static struct mtd_partition cxe_partitions[NUM_PARTITIONS] = {
    { .name = "CXE-2130 flash boot partition", 
      .offset = 0, 
      .size =   BOOT_PARTITION_SIZE_KiB*1024 },
    { .name = "CXE-2130 kernel partition", 
      .offset = BOOT_PARTITION_SIZE_KiB*1024, 
      .size = (KERN_PARTITION_SIZE_KiB)*1024 },
    { .name = "CXE-2130 application partition", 
      .offset = (BOOT_PARTITION_SIZE_KiB+KERN_PARTITION_SIZE_KiB)*1024, 
      .size = (APP_PARTITION_SIZE_KiB)*1024 },
    { .name = "CXE-2130 data partition", 
      .offset = (BOOT_PARTITION_SIZE_KiB +
		 KERN_PARTITION_SIZE_KiB +
		 APP_PARTITION_SIZE_KiB) * 1024, 
      .size = (DATA_PARTITION_SIZE_KiB)*1024 },
    { .name = "CXE-2130 eeprom partition", 
      .offset = (BOOT_PARTITION_SIZE_KiB +
		 KERN_PARTITION_SIZE_KiB +
		 APP_PARTITION_SIZE_KiB +
		 DATA_PARTITION_SIZE_KiB) * 1024 }
};




static void flash_write_protection_off(void)
{
    volatile immap_t *immr = (immap_t *)CFG_IMMR;

    /* Set bit PA13 to high */
    immr->im_ioport.iop_padat |=  (0x0004);
    immr->im_ioport.iop_paodr &= ~(0x0004);
    immr->im_ioport.iop_papar &= ~(0x0004);
    immr->im_ioport.iop_padir |=  (0x0004);
}

static void flash_write_protection_on(void)
{
    volatile immap_t *immr = (immap_t *)CFG_IMMR;

    /* Set bit PA13 to low */
    immr->im_ioport.iop_padat &= ~(0x0004);
    immr->im_ioport.iop_paodr &= ~(0x0004);
    immr->im_ioport.iop_papar &= ~(0x0004);
    immr->im_ioport.iop_padir |=  (0x0004);
}



/*
************************************************************
** Enable/Disable VPP to write to flash
************************************************************
*/

static spinlock_t dnpc_spin   = SPIN_LOCK_UNLOCKED;
static int        vpp_counter = 0;
/*
** This is what has to be done for the DNP board ..
*/
static void cxe_2130_set_vpp(struct map_info *not_used, int on)
{
	spin_lock_irq(&dnpc_spin);

	if (on)
	{
		if(++vpp_counter == 1)
			flash_write_protection_off();
	}
	else
	{
		if(--vpp_counter == 0)
			flash_write_protection_on();
		else if(vpp_counter < 0)
			vpp_counter = 0;
	}
	spin_unlock_irq(&dnpc_spin);
}



int __init init_cxe_2130(void)
{
	printk(KERN_NOTICE "CXE-2130 flash device: %x at %x\n", WINDOW_SIZE*4, WINDOW_ADDR);
	cxe_2130_map.virt = (unsigned long)ioremap(WINDOW_ADDR, WINDOW_SIZE * 4);

	if (!cxe_2130_map.virt) {
		printk("Failed to ioremap\n");
		return -EIO;
	}
	simple_map_init(&cxe_2130_map);
	mymtd = do_map_probe("cfi_probe", &cxe_2130_map);
	if (mymtd) {
		mymtd->owner = THIS_MODULE;
		add_mtd_partitions(mymtd, cxe_partitions, NUM_PARTITIONS );
		return 0;
	}

	iounmap((void *)cxe_2130_map.virt);
	return -ENXIO;
}

static void __exit cleanup_cxe_2130(void)
{
	if (mymtd) {
		del_mtd_device(mymtd);
		map_destroy(mymtd);
	}
	if (cxe_2130_map.virt) {
		iounmap((void *)cxe_2130_map.virt);
		cxe_2130_map.virt = 0;
	}
}

module_init(init_cxe_2130);
module_exit(cleanup_cxe_2130);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("LVL7 Systems.");
MODULE_DESCRIPTION("MTD map driver CXE-2130 boards");
