/*
 * (C) Copyright 2003
 * John W. Linville, LVL7 Systems, Inc., linville@lvl7.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>

void reset64();

typedef enum
{
  FALSE = 0,
  TRUE
} bool;

#define RAM_TEST_VAL  0xAAAAAAAA

int checkboard (void)
{

	printf("Board: DLink GSW7224TG \n");

	return 0;
}


/* 
  DESCR:  return TRUE if two ram locations are aliased
  PARMS:  pRam1 - ptr to 1st ram location
          pRam2 - ptr to 2nd ram location
  RETURN: TRUE, if location 1 aliases with location 2
  PRE:    - pRamN non-null
 */
static bool ramIsAliased(volatile uint* pRam1, volatile uint* pRam2)
{
  bool retVal = FALSE;
  
  /* write test value to 1st location */
  *pRam1 = RAM_TEST_VAL;

  /* if value aliases in 2nd location, check whether inverted value aliases */
  if (*pRam2 == RAM_TEST_VAL)
  {
    *pRam1 = ~RAM_TEST_VAL;
    retVal = (*pRam2 == ~RAM_TEST_VAL);
  }

  return retVal;
}


/*
  DESCR:  discover size of installed DRAM, up to CFG_MAX_RAM_SIZE
          - contrary to the function name, no DRAM init is done here
  PARMS:  board_type - unused
  RETURN: size of DRAM, in bytes
  PRE:    - ram is int width (32-bit wide)
          - max possible DRAM size, CFG_MAX_RAM_SIZE, is a power-of-2
          - actual discovered DRAM size is power-of-2 (ie, 16M, 32M, etc.)
  POST:   - SDRAM_BASE[0, 1, 2, 4, ..., (CFG_MAX_RAM_SIZE/2)] are accessed
 */
long int initdram (int board_type)
{
	volatile uint* const pRamOrg = (void*) CFG_SDRAM_BASE;
	volatile uint* const pTermOrg = (void*) (CFG_SDRAM_BASE + CFG_MAX_RAM_SIZE);
	volatile uint* pRam = pRamOrg;
	long int lRamSize;

  /*
    while (ram ptr < ram boundary) 
          AND (ram location is volatile)
          AND (ram location is not aliased with origin)
    NOTE:   - ram is volatile if it can alias with itself (oooo ... clever)
   */
  while (
         (pRam < pTermOrg)
         && ramIsAliased(pRam, pRam)
         && ((pRam == pRamOrg) || !ramIsAliased(pRam, pRamOrg))
        )
  {
    /*
      adjust loop var:
      - set pRam to first word of next power-of-2 RAM boundary
     */
    if (pRam == pRamOrg)
    {
      pRam += 1;
    }
    else
    {
      pRam = pRamOrg + ((pRam - pRamOrg) << 1);
    }
  }

  /* return byte count between first invalid location and origin of DRAM */
  /* return (pRam - pRamOrg) << 2; */
  lRamSize = (pRam - pRamOrg) << 2;
  
  /* If memory detected is less than 64MB then it means the memctrl
   * col size has to be flipped to 512 as the board must have only 
   * 64MB. Hence call the alternate reset that sets the colsize to 512.
   */
  if (lRamSize < 0x4000000)
  {
    printf ("Less than 128MB hence resetting to try 64MB\n");
    reset64() ;
  }
  return (lRamSize);
}

/* Turns on Green Status LED */
void boot_status_green_led_on(void)
{
    ulong reg;
    
    reg = *((volatile ulong *)0xb8000064);
    
    reg |= 0x00000040;
    
	/* Preset GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000064) = reg;
}

/* Turns off Green Status LED */
void boot_status_green_led_off(void)
{
    ulong reg;
    
    reg = *((volatile ulong *)0xb8000064);
    
    reg &= ~0x00000040;
    
	/* Preset GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000064) = reg;
}

/* Turns on Yellow Status LED */
void boot_status_yellow_led_on(void)
{
    ulong reg;
    
    reg = *((volatile ulong *)0xb8000064);
    
    reg |= 0x00000080;
    
	/* Preset GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000064) = reg;
}

/* Turns off Yellow Status LED */
void boot_status_yellow_led_off(void)
{
    ulong reg;
    
    reg = *((volatile ulong *)0xb8000064);
    
    reg &= ~0x00000080;
    
	/* Preset GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000064) = reg;
}

void show_boot_progress (int status)
{
    if (status < 0) { /* Error */
        boot_status_green_led_off();
        boot_status_yellow_led_on();
        return;
    }
    switch (status) {
        case 97:    /* Yellow LED On */
            boot_status_yellow_led_on();
            boot_status_green_led_off();
            break;
        case 98:    /* Green and Yellow LED Off */
            boot_status_yellow_led_off();
            boot_status_green_led_off();
            break;
        case 99:    /* Green LED On */
            boot_status_yellow_led_off();
            boot_status_green_led_on();
            break;
        default:    /* LED Off */
            boot_status_yellow_led_off();
            boot_status_green_led_off();
            break;
    }
}

int board_pre_init(void)
{
	/* Enable CS01 (aka M_CS1) */
	*((volatile ulong *)0xb8000100) = 0x00000001;

	/* Set timings for CS01 */
	*((volatile ulong *)0xb8000104) = 0x0100010a;

	/* Enable CS23 (aka M_CS2) */
	*((volatile ulong *)0xb8000110) = 0x00000001;

	/* Set timings for CS23 */
	*((volatile ulong *)0xb8000114) = 0x0100010a;

	/* Enable CS4  (aka M_CS3) */
	*((volatile ulong *)0xb8000120) = 0x00000001;

	/* Set timings for CS4 */
	*((volatile ulong *)0xb8000124) = 0x0100010a;

	/* Set timings for FLASH_CS */
	*((volatile ulong *)0xb800012c) = 0x0100010a;

	/* Set divider for UART clock */
	*((volatile ulong *)0xb80000a4) = 0x00000036;

	/* Set SBIPSFlag for MIPS core */
	*((volatile ulong *)0xb8005f08) = 0x043f0201;

	/* Set SBIntVec for MIPS core */
	*((volatile ulong *)0xb8005f94) = 0x00000001;

	/* Preset GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000064) = 0x00000083;
	
	/* Enable GPIO outputs for reset and LED control */
	*((volatile ulong *)0xb8000068) = 0x000000c3;
	
	return (0);
}
